/**
 * @file
 * Provides Splide extensions for onwheel event.
 */

(function ($, _ds, _win, _doc) {

  'use strict';

  var xZoom = function (Splide, Components) {
    var EL_SPLIDE_ROOT = Splide.root;
    var OPT = Splide.options;
    var OZ = OPT.zoom || {};
    var V_MAX = OZ.max || 1.5;
    var V_MIN = OZ.min || 1;
    var IS_ENABLED = OZ.on || false;
    var IS_SCALABLE = OZ.scale || false;
    var IS_CLICKABLE = OZ.click || false;
    var S_ZOOM_ROOT = OZ.root || false;
    var C_NICK = OZ.nickClass || 'sbox';
    var C_IS_NICK = 'is-' + C_NICK;
    var C_ZOOM_ROOT = C_IS_NICK + '-zoomed';
    var C_ZOOM_XL = C_IS_NICK + '-zoom-xl';
    var S_ZOOM_PARENT = OZ.target || '.slide__media';
    var C_PINCH = C_IS_NICK + '-zoom';
    var C_DRAG = C_IS_NICK + '-dragging';
    var C_TARGET = C_IS_NICK + '-moved';
    var MIN_DEVICE_WIDTH = OZ.minWidth || '768px';
    var USE_TRANSFORM = OZ.transform || false;
    var IS_PINCHED = false;
    var C_IS_ZOOMABLE = 'is-zoomable';
    var S_IS_ZOOMABLE = '.' + C_IS_ZOOMABLE;
    var E_DBLCLICK = 'dblclick';
    var V_POS = {
      x: 0,
      y: 0
    };
    var EL_ZOOM_ROOT = null;
    var EL_ZOOM_PARENT = null;
    var EL_ZOOMED = null;
    var V_OFFSET = 0;
    var V_SIZES = {};
    var V_WIDTH;
    var V_HEIGHT;
    var EL_ZOOM_PARENTS = [];
    var V_WINSIZE = $.windowSize() || {};
    var V_TOGGLING;
    var V_EASING = 'cubic-bezier(.4,0,.22,1)';

    function dim(el, which) {
      return which === 'width' ? el.offsetWidth : el.offsetHeight;
    }

    function drag(v) {
      Splide.options = {
        drag: v
      };
    }

    // https://stackoverflow.com/questions/3971841
    function aspectRatio(srcWidth, srcHeight, maxWidth, maxHeight) {
      var ratio = Math.min(maxWidth / srcWidth, maxHeight / srcHeight);

      return {
        width: srcWidth * ratio,
        height: srcHeight * ratio,
        ratio: ratio
      };
    }

    function delayedClass(cn, v, remove) {
      $.addClass(cn, C_IS_NICK + '-' + v);

      _win.clearTimeout(V_TOGGLING);

      V_TOGGLING = _win.setTimeout(function () {
        V_TOGGLING = null;

        if ($.isUnd(remove)) {
          $.removeClass(cn, C_IS_NICK + '-' + v);
        }
      }, 250);
    }

    function toTransform(x, y, scale) {
      var str = 'translate3d(' + x + 'px, ' + (y || 0) + 'px, 0)';

      if (!$.isUnd(scale)) {
        str += ' scale3d(' + scale + ', ' + scale + ', 1)';
      }

      return str;
    }

    function inject(el, x, y, scale, transform) {
      var tg = USE_TRANSFORM ? $.closest(el, S_ZOOM_PARENT) : el;

      if (!tg) {
        return;
      }

      if (USE_TRANSFORM) {
        tg = $.closest(el, S_ZOOM_PARENT);
        tg.style.transform = toTransform(x, y, scale);
      }
      else {
        tg.style.left = x === '' ? '' : x + 'px';
        tg.style.top = y === '' ? '' : y + 'px';
      }

      $.on(tg, 'transitionend.sz', function () {
        tg.style.transition = '';
      });

      _win.setTimeout(function () {
        tg.style.transition = '';
      }, 500);
    }

    function updateDim(el, fit) {
      if (!el || !el.dataset) {
        return;
      }
      var ow = el.dataset.sNw;
      var oh = el.dataset.sNh;

      var cw = el.dataset.sCw;
      var ch = el.dataset.sCh;

      var eh = $.attr(el, 'height', oh);
      var ew = $.attr(el, 'width', ow);

      V_WIDTH = fit ? ew : cw;
      V_HEIGHT = fit ? eh : ch;

      el.style.width = V_WIDTH + 'px';
      // el.style.height = (fit ? eh : ch) + 'px';

      var picture = $.closest(el, 'picture');
      if ($.isElm(picture)) {
        picture.style.width = V_WIDTH + 'px';
        picture.style.height = V_HEIGHT + 'px';
      }

      var xl = $.toInt(oh, 0) >= V_WINSIZE.height;
      var tg = $.closest(el, S_ZOOM_PARENT);

      if (tg) {
        $[xl ? 'addClass' : 'removeClass'](tg, C_ZOOM_XL);
      }
    }

    return {
      currSlide: null,
      canZoom: true,
      sizes: function (img) {
        V_SIZES = _ds.checkSizes(img || EL_ZOOMED, EL_ZOOM_PARENT);
        return V_SIZES;
      },

      isZoomed: function () {
        return IS_ENABLED && $.hasClass(this.zoomRoot(), C_ZOOM_ROOT);
      },

      toggleClass: function (zooming, cls) {
        $[zooming ? 'addClass' : 'removeClass'](this.zoomRoot(), cls || C_ZOOM_ROOT);
      },

      on: function () {
        this.canZoom = true;
      },

      off: function () {
        this.canZoom = false;
      },

      onMounted: function () {
        var me = this;

        EL_ZOOM_PARENTS = $.findAll(EL_SPLIDE_ROOT, S_ZOOM_PARENT);

        if (EL_ZOOM_PARENTS.length) {
          $.each(EL_ZOOM_PARENTS, function (el) {
            var img = $.find(el, S_IS_ZOOMABLE);

            if ($.isElm(img)) {
              if ($.isDecoded(img)) {
                me.fit(img);
              }
              else {
                $.one(img, 'load.sz', me.fit.bind(me));
              }
            }
          });
        }
      },

      onActive: function (slide) {
        var me = this;

        me.currSlide = slide;
        EL_ZOOMED = $.find(slide.slide, S_IS_ZOOMABLE);

        if (me.currSlide) {
          EL_ZOOM_PARENT = $.find(me.currSlide.slide, S_ZOOM_PARENT);
          if ($.isElm(EL_ZOOM_PARENT)) {
            EL_ZOOMED = $.find(EL_ZOOM_PARENT, S_IS_ZOOMABLE);

            if (EL_ZOOMED && EL_ZOOMED.dataset) {
              var fit = $.toInt(EL_ZOOMED.dataset.sFit, 0) === 1;

              if (fit) {
                $.addClass(EL_ZOOM_PARENT, 'is-zoom-fit');
              }
            }
          }

          me.on();
        }
      },

      onWheel: function (e, data) {
        var me = this;
        var el = e.target;

        if ($.matchMedia(MIN_DEVICE_WIDTH, 'min')) {
          if (me.isZoomed()) {
            var cn = me.zoomRoot();

            $.addClass(cn, C_IS_NICK + '-wheeled');

            inject(el, data.x, data.y, data.scale || V_MAX);
          }
          else {
            me.canZoom = data.dir === 'down';
            me.zoomon(el, true);
          }
        }
      },

      mount: function () {
        var me = this;

        if (!IS_ENABLED) {
          return;
        }

        Splide.on('mounted.spz', function () {
          setTimeout(function () {
            me.onMounted();
          }, 300);
        });

        Splide.on('active.spz', me.onActive.bind(me));

        Splide.on('inactive.spz', function (slide) {
          var oldImg = $.find(slide.slide, S_IS_ZOOMABLE);

          if ($.isElm(oldImg)) {
            me.zoomOut(oldImg);
          }
        });

        // Must be outside splide events to support destroyed splide.
        me.dragon();
      },

      fit: function (el) {
        var img = el.target || el;
        var p = $.closest(img, '.slide');
        var ph = dim(p, 'height');
        var pw = ph;
        var cw;
        var ch;
        var ratio = 0;
        var ow = img.naturalWidth;
        var oh = img.naturalHeight;
        var eh = $.attr(img, 'width', oh);
        var nh = $.toInt(oh, 0);
        var nw = $.toInt(ow, 0);

        img.style.width = 'auto';
        // img.style.height = 'auto';

        if (ow > pw && ow > oh) {
          ratio = ow / oh;
          cw = pw;
          ch = (pw / ratio);
        }
        else if (oh > ph && oh > ow) {
          ratio = oh / ow;

          cw = (ph / ratio);
          ch = ph;
        }
        else {
          cw = pw;
          ch = ph;
        }

        var fit = eh < ph;

        if (!img.dataset) {
          return;
        }

        img.dataset.sFit = fit ? 1 : 0;

        img.dataset.sNw = nw;
        img.dataset.sNh = nh;

        img.dataset.sCw = $.toInt(cw, 0);
        img.dataset.sCh = $.toInt(ch, 0);

        img.dataset.sRatio = $.toInt(ratio, 0);

        updateDim(img, fit);
        if (nh < V_WINSIZE.height && nw < V_WINSIZE.width) {
          $.addClass(p, 'is-slide-sm');
        }
      },

      zoomRoot: function () {
        if (!$.isElm(EL_ZOOM_ROOT) && S_ZOOM_ROOT) {
          EL_ZOOM_ROOT = typeof S_ZOOM_ROOT === 'string' ? $.find(_doc, S_ZOOM_ROOT) : S_ZOOM_ROOT;
        }
        return $.isElm(EL_ZOOM_ROOT) ? EL_ZOOM_ROOT : EL_SPLIDE_ROOT;
      },

      zoomIn: function (el) {
        var me = this;
        var valid = el && el.dataset;
        var nh = $.toInt(el.dataset.sNh, 0) || el.naturalHeight;
        var nw = $.toInt(el.dataset.sNw, 0) || el.naturalWidth;
        var fit = valid ? $.toInt(el.dataset.sFit, 0) === 1 : false;
        var small = nh < V_WINSIZE.height && nw < V_WINSIZE.width;
        var largeEnough = nw > $.toInt(el.dataset.sCw, 0) && nh > $.toInt(el.dataset.sCh, 0);

        if (small || fit || !IS_SCALABLE) {
          me.toggleClass(false);

          if (small && largeEnough) {
            // @todo.
          }
          else {
            return false;
          }
        }

        me.toggleClass(true);

        if (valid) {
          var slide = $.closest(el, '.slide');

          $.addClass(slide, C_IS_ZOOMABLE + '-slide');

          /*
          V_MAX = Math.min(
            nw / $.toInt(el.dataset.sCw, 0),
            nh / $.toInt(el.dataset.sCh, 0)
          );

          V_MAX = V_MAX > 2 ? 2 : V_MAX;
          */

          var limit = V_WINSIZE.width > V_WINSIZE.height ? V_WINSIZE.width : V_WINSIZE.height;

          // if ($.matchMedia(MIN_DEVICE_WIDTH, 'min')) {
          var ratio = aspectRatio(nw, nh, limit - 80, nh - 80);
          el.style.width = ratio.width + 'px';
          el.style.height = ratio.height + 'px';

          delayedClass(me.zoomRoot(), 'zoomin');
          // }
        }

        setTimeout(function () {
          me.off();
        }, 250);

        drag(false);
      },

      zoomOut: function (el) {
        var me = this;

        if (el) {
          var cn = $.closest(el, S_ZOOM_PARENT);
          var w = $.toInt(el.dataset.sCw, 0) || V_WIDTH;

          if (USE_TRANSFORM) {
            cn.style.transform = 'scale3d(' + V_MIN + ', ' + V_MIN + ', ' + V_MIN + ')';
            cn.style.transition = 'transform 300ms ' + V_EASING;
          }
          else {
            el.style.transition = 'width 300ms ' + V_EASING;
            el.style.left = '';
            el.style.top = '';
          }

          el.style.width = w + 'px';
          el.style.height = '';

          delayedClass(me.zoomRoot(), 'zoomout');

          me.toggleClass(false);
        }

        setTimeout(function () {
          me.on();
          drag(true);
        }, 250);
      },

      zoomon: function (el, reset) {
        var me = this;
        el = el || EL_ZOOMED;
        var img = $.hasClass(el, C_IS_ZOOMABLE) ? el : $.find(el, S_IS_ZOOMABLE);

        if (me.canZoom) {
          me.zoomIn(img);
        }
        else {
          me.zoomOut(img);
          if (reset) {
            inject(img, '', '', 1);
          }
        }
      },

      onClick: function (e, data) {
        var me = this;
        var img = e.target;

        if ($.hasClass(img, C_IS_ZOOMABLE)) {
          EL_ZOOMED = img;

          if (IS_PINCHED) {
            me.zoomOut(img);
            IS_PINCHED = false;
          }
          else if (_ds.doubletap() || e.type === E_DBLCLICK) {
            // e.preventDefault();
            me.zoomon(img);
          }
        }
      },

      onZoomIn: function (e, data) {
        var me = this;
        var img = e.target;

        if ($.hasClass(img, C_IS_ZOOMABLE)) {
          EL_ZOOMED = img;

          // me.zoomon(img);
          me.toggleClass(true);
        }
      },

      onZoomOut: function (e, data) {
        var me = this;
        var img = e.target;

        if ($.hasClass(img, C_IS_ZOOMABLE)) {
          EL_ZOOMED = img;

          me.zoomOut(img);
        }
      },

      onPinched: function (e, data) {
        IS_PINCHED = true;
      },

      dragon: function () {
        var me = this;
        var cn = me.zoomRoot();
        var phase;
        var dir;
        var reset = false;
        var opts = OZ;
        var el = EL_ZOOM_PARENT;
        var sy = 0;

        function start(data) {
          el = el || data.el;
          var img = data.img;
          var tg = USE_TRANSFORM ? el : img;

          tg.style.transition = '';

          $.removeClass(me.zoomRoot(), C_IS_NICK + '-wheeled');
        }

        function move(data) {
          el = data.el || el;
          var img = data.img;
          var touchData = data.zoom;
          var inoutCls = C_PINCH +
            'in ' + C_PINCH + 'out';

          me.toggleClass(false, inoutCls);

          if (me.isZoomed()) {
            if (touchData) {
              me.toggleClass(true, C_PINCH + touchData);
            }
            else {
              me.toggleClass(false, inoutCls);

              // if ($.matchMedia(MIN_DEVICE_WIDTH, 'min')) {
              inject(img, data.x, data.y, data.scale || V_MAX);
              // img.style.transition = '';
              // }
            }
          }
          else {
            // me.zoomOut(img);
            me.toggleClass(false, inoutCls);
          }
        }

        function end(data, sy) {
          el = data.el || el;
          dir = data.dir;
          V_POS.x = data.x;
          V_POS.y = data.y;

          var img = data.img;

          me.sizes(img);
          sy = Math.ceil((V_SIZES.h - V_SIZES.ph) / 2);
          var hw = V_SIZES.nw / 2;
          var hh = V_WINSIZE.height / 2;
          // var bh = data.rect.bottom / 2;
          var downward = dir === 'down';
          // var upward = dir === 'up';

          V_OFFSET = Math.abs(data.y) >= sy;

          var toRight = dir === 'right';
          var toLeft = dir === 'left';
          var horiz = toRight || toLeft;

          if (horiz) {
            V_POS.x = toRight ? 0 : -1;
            reset = Math.abs(data.x) > hw;
          }
          else if (V_OFFSET) {
            // dir === 'up' && V_POS.y < 0
            V_POS.y = downward && data.y > 0 ? sy : -sy;
            reset = downward && Math.abs(data.y) > hh; // || upward && Math.abs(data.y) > bh;
          }

          setTimeout(function () {
            if (reset && data.trigger === data.event.type) {
              var prop = USE_TRANSFORM ? 'transform' : 'left';
              var tg = USE_TRANSFORM ? el : img;

              tg.style.transition = prop + ' 300ms ' + V_EASING;

              if (!USE_TRANSFORM) {
                V_POS.x = '';
                V_POS.y = '';
              }

              inject(img, V_POS.x, V_POS.y, data.scale || V_MAX);

              /*
              $.on(img, 'transitionend.sz', function () {
                img.style.transition = '';
              });

              _win.setTimeout(function () {
                img.style.transition = '';
              }, 500);
              */
            }
          }, 600);
        }

        var callback = function (e, data) {
          el = data.el || el;
          dir = data.dir;
          phase = data.phase;

          data.img = $.find(el, S_IS_ZOOMABLE);
          me.sizes(data.img);

          sy = Math.ceil((V_SIZES.h - V_SIZES.ph) / 2);

          if (phase === 'start') {
            start(data);
          }
          else if (phase === 'move') {
            move(data);
          }
          else if (phase === 'end') {
            end(data, sy);
          }
        };

        opts.onClick = IS_CLICKABLE ? me.onClick.bind(me) : false;
        opts.onZoomIn = me.onZoomIn.bind(me);
        opts.onZoomOut = me.onZoomOut.bind(me);
        opts.onPinched = me.onPinched.bind(me);
        opts.onWheel = me.onWheel.bind(me);
        opts.callback = callback;
        opts.dragClass = C_DRAG;
        opts.activatedClass = C_ZOOM_ROOT;
        opts.zoomClass = C_IS_ZOOMABLE;
        opts.targetClass = C_TARGET;
        opts.max = V_MAX;

        var items = [];
        $.each(Components.Elements.slides, function (slide) {
          var item = $.find(slide, S_ZOOM_PARENT);

          if ($.isElm(item)) {
            items.push(item);
          }
        });

        if (items.length) {
          opts.elms = items;

          new SwipeDetect(cn, opts);
        }
      }
    };
  };

  _ds.extend({
    xZoom: xZoom
  });

})(dBlazy, dSplide, this, this.document);
